/* led.h */

#include "timer.h"

typedef unsigned int led_pattern;

extern led_pattern led_pattern_active[];
extern led_pattern led_pattern_unknown[];

void led_init(void);
void led_set(bool on);
void led_update(void);
void led_set_pattern(led_pattern *pattern);
