/* motor.h */

extern float temp_kp;

void motor_pid_update(float troll, float mroll,
		      float tpitch, float mpitch,
		      float tyaw, float myaw);
void motor_kill(void);
void motor_set_throttle(float t);
