/* dcm.h */

#include "types.h"

extern float delta_t;

void dcm_update(float omega_x, float omega_y, float omega_z);
void dcm_normalise(void);
bool dcm_renormalise(float *v);
void dcm_dump(void);
void dcm_send_packet(void);
void dcm_setvector(float x, float y, float z);
void dcm_drift_correction(float x, float y, float z);
void dcm_attitude_error(float x, float y, float z);
