/* panic.h */

/* OMG we're all going to die!!!! */

void panic(unsigned int reason);

/* Panic code goes in the low 8 bits */
#define PANIC_WATCHDOG_TIMEOUT 0x100
#define PANIC_SENSOR_FAIL 0x200

#ifdef PANIC_CHECKPOINT
extern unsigned int checkpoint;
#define CHECKPOINT(x) do { checkpoint = (x); } while (0)
#else
#define CHECKPOINT(x)
#endif
