/* watchdog.h */

void watchdog_kick(unsigned int module);
void watchdog_check(void);
void init_watchdog(void);

#define WATCHDOG_STICK 0
#define WATCHDOG_GYRO  1
#define WATCHDOG_ACCEL 2

#define WATCHDOG_MODULES 3
